@@ -44,6 +44,10 @@
 
 #define d(x)
 
+typedef struct _YelpHtmlBoxNode YelpHtmlBoxNode;
+
+#define YELP_HTML_BOX_NODE(x)  ((YelpHtmlBoxNode *)(x))
+
 struct _YelpHtmlPriv {
     HtmlView        *view;
 
@@ -51,8 +55,19 @@ struct _YelpHtmlPriv {
     YelpURI         *base_uri;
 
     DomNodeIterator *find_iter;
-    DomNode         *find_node;
     gint             find_offset;
+
+    gchar           *find_str;
+
+    GList           *find_list;
+    GList           *find_elem;
+    gboolean         find_is_forward;
+};
+
+struct _YelpHtmlBoxNode {
+    HtmlBox *find_box;
+    glong    find_index;
+    gint     find_len;
 };
 
 static void      html_init               (YelpHtml           *html);
@@ -72,7 +87,21 @@ static void      html_link_clicked_cb    (HtmlDocument       *doc,
 static void      html_title_changed_cb   (HtmlDocument       *doc,
 					  const gchar        *new_title,
 					  YelpHtml           *html);
+
+static DomNode * html_get_dom_node       (HtmlDocument       *doc,
+					  const gchar        *node_name);
+static void      html_clear_box_text     (gpointer            data, 
+                                          gpointer            user_data);
+static gint      html_cmp_box_text       (gconstpointer       a,
+		                          gconstpointer       b);
+static void      html_get_box_list       (YelpHtml           *html);
+
 static void      html_clear_find_data    (YelpHtml           *html);
+static gboolean  html_find_text          (YelpHtml           *html,
+					  const gchar *       find_string,
+					  gboolean            match_case,
+					  gboolean            wrap,
+					  gboolean            forward);
 
 #define BUFFER_SIZE 16384
 
@@ -121,6 +150,8 @@ html_init (YelpHtml *html)
     priv->doc         = html_document_new ();
     priv->base_uri    = NULL;
 
+    priv->find_is_forward = TRUE;
+
     html_view_set_document (HTML_VIEW (priv->view), priv->doc);
 
     g_signal_connect (G_OBJECT (priv->doc), "link_clicked",
@@ -314,6 +345,7 @@ yelp_html_clear (YelpHtml *html)
     YelpHtmlPriv *priv;
 
     g_return_if_fail (YELP_IS_HTML (html));
+    html_clear_find_data (html);
 
     priv = html->priv;
 
@@ -391,13 +423,336 @@ html_clear_find_data (YelpHtml *html)
 
     /* Reset find data when we load a new page. */
     if (priv->find_iter) {
+ 	html_selection_clear (priv->view);
+
+	if (priv->find_list) {
+	    g_list_foreach (priv->find_list, html_clear_box_text, NULL);
+	    
+	    g_list_free (priv->find_list);
+
+	    priv->find_list = NULL;
+	}
+
+	g_free (priv->find_str);
+	priv->find_str = NULL;
+
+	priv->find_elem = NULL;
+
 	g_object_unref (priv->find_iter);
 	priv->find_iter = NULL;
     }
 }
 
-/* This code is really ugly, need to clean up. */
-void
+static DomNode *
+html_get_dom_node (HtmlDocument *doc,
+		   const gchar  *node_name)
+{
+    DomNodeIterator *iter;
+    DomNode         *node;
+
+    g_return_val_if_fail (HTML_IS_DOCUMENT (doc), NULL);
+    g_return_val_if_fail (node_name != NULL, NULL);
+
+    iter = dom_DocumentTraversal_createNodeIterator
+	(DOM_DOCUMENT_TRAVERSAL (doc->dom_document),
+	 DOM_NODE (doc->dom_document),
+	 DOM_SHOW_ALL,
+	 NULL,
+	 FALSE,
+	 NULL);
+	
+    while ((node = dom_NodeIterator_nextNode (iter, NULL))) {
+	if (!g_ascii_strcasecmp (dom_Node__get_nodeName (node), node_name)) {
+	    break;
+	}
+    }
+	
+    g_object_unref (iter);
+
+    return node;
+}
+
+static void
+html_clear_box_text (gpointer data, gpointer user_data)
+{
+    g_free (data);
+}
+
+static gint
+html_cmp_box_text (gconstpointer a,
+		   gconstpointer b)
+{
+    YelpHtmlBoxNode *box_node;
+    HtmlBox         *html_box;
+
+    box_node = YELP_HTML_BOX_NODE (a);
+    html_box = HTML_BOX (b);
+
+    return ! (html_box == box_node->find_box);
+}
+
+static void 
+html_get_box_list (YelpHtml *html)
+{
+    YelpHtmlPriv    *priv;
+    YelpHtmlBoxNode *box_node;
+    DomNode         *node;
+    HtmlBox         *html_box;
+    gint             len;
+    glong            total_len;
+
+    g_return_if_fail (YELP_IS_HTML (html));
+
+    priv = html->priv;
+
+    g_return_if_fail (priv->find_iter != NULL);
+    g_assert (priv->find_list == NULL);
+
+    node = dom_NodeIterator_nextNode (priv->find_iter, 
+				      NULL);
+
+    priv->find_list = NULL;
+
+    while (node) {
+	html_box = html_view_find_layout_box (priv->view,
+					      node, FALSE);
+	total_len = 0;
+	while (html_box) {
+	    if (HTML_IS_BOX_TEXT (html_box) &&
+		(!g_list_find_custom (priv->find_list,
+				      html_box,
+				      html_cmp_box_text))) {
+		html_box_text_get_text (HTML_BOX_TEXT (html_box), &len);
+
+		if (len > 0) {
+		    box_node             = g_new0 (YelpHtmlBoxNode, 1);
+		    
+		    box_node->find_box   = html_box;
+		    box_node->find_index = total_len;
+		    box_node->find_len   = len;
+		    
+		    priv->find_list      = g_list_append (priv->find_list, box_node);
+
+		    total_len += len;
+		}
+	    } else {
+		total_len = 0;
+	    }
+				
+	    if (html_box->children)
+		html_box = html_box->children;
+	    else if (html_box->next)
+		html_box = html_box->next;
+	    else
+		html_box = NULL;
+	}
+			
+	node = dom_NodeIterator_nextNode (priv->find_iter,
+					  NULL);
+    }    
+}
+
+static void
+html_find_selection_set (HtmlView *view,
+			 DomNode  *node,
+			 gint      offset,
+			 gint      len)
+{
+    g_return_if_fail (HTML_IS_VIEW (view));
+    g_return_if_fail (DOM_IS_NODE (node));
+    g_return_if_fail (offset >= 0);
+    g_return_if_fail (len >= 0);
+
+    html_selection_set (view,
+			node,
+			offset,
+			len);
+
+    html_view_scroll_to_node (view,
+			      node,
+			      HTML_VIEW_SCROLL_TO_TOP);    
+}
+
+static void
+html_find_reset (YelpHtml *html)
+{
+    
+    YelpHtmlPriv *priv;
+
+    g_return_if_fail (YELP_IS_HTML (html));
+
+    priv = html->priv;
+
+    g_return_if_fail (priv->find_list != NULL);
+
+    priv->find_elem       = priv->find_list;
+    priv->find_offset     = 0;
+    priv->find_is_forward = TRUE;    
+
+    g_free (priv->find_str);
+    priv->find_str = NULL;
+}
+			 
+static gboolean
+html_find_text (YelpHtml    *html,
+		const gchar *find_string,
+		gboolean     match_case,
+		gboolean     wrap,
+		gboolean     forward)		
+{
+    YelpHtmlPriv    *priv;
+    YelpHtmlBoxNode *box_node;
+    gchar           *box_text;
+    gchar           *str;
+    gchar           *haystack;
+    gchar           *hit;
+    gchar           *found_str;
+    GList           *next_elem;
+    GList           *next_list;
+    GList           *first_elem;
+    gint             len;
+    gint             next_offset;
+    gboolean         found;
+
+    g_return_val_if_fail (YELP_IS_HTML (html), FALSE);
+    g_return_val_if_fail (find_string != NULL, FALSE);
+
+    priv  = html->priv;
+
+    next_list   = priv->find_list;
+    next_elem   = priv->find_elem;
+    next_offset = priv->find_offset;
+    
+    if (priv->find_is_forward != forward) {
+	next_list = g_list_reverse (next_list);
+
+	if (next_elem) {
+	    box_node = next_elem->data;
+
+	    if (priv->find_str) {
+		next_offset = box_node->find_len - next_offset + strlen (priv->find_str);
+	    } else {
+		if (!forward) {
+		    next_offset = box_node->find_len;
+		} else {
+		    g_assert (next_offset == 0);
+		}
+	    }
+	}
+    } 
+
+    first_elem = next_elem;
+    found      = FALSE;
+    found_str  = NULL;
+
+    while (next_elem) {
+	box_node = next_elem->data;
+	box_text = html_box_text_get_text (HTML_BOX_TEXT (box_node->find_box), &len);
+
+	g_assert (box_node->find_len == len);
+
+	if (len > 0) {
+	    str = g_strndup (box_text, len);
+
+	    if (!match_case) {
+		haystack = g_utf8_casefold (str, -1);
+	    } else {
+		haystack = g_strdup (str);
+	    }
+
+	    if (forward) {
+		hit = strstr (haystack + next_offset, find_string);
+	    } else {
+		hit = g_strrstr_len (haystack, len - next_offset,
+				     find_string);
+	    }
+
+	    if (hit) {
+		html_find_selection_set (priv->view,
+					 box_node->find_box->dom_node,
+					 box_node->find_index + (hit - haystack),
+					 strlen (find_string));
+		found = TRUE;
+
+		if (forward) {
+		    next_offset = (hit - haystack + strlen (find_string));
+		}else {
+		    next_offset = strlen (hit);
+		}
+		
+		found_str = g_strndup (box_text + (hit - haystack), strlen (find_string));
+	    }
+	    
+	    g_free (str);
+	    g_free (haystack);
+	}
+
+	if (found)
+	    break;
+		
+	next_offset = 0;
+	next_elem   = g_list_next (next_elem);
+
+	if (first_elem == next_elem) {
+	    next_elem = NULL;
+	} else if (!next_elem) {
+	    if (wrap) {
+		next_elem  = next_list;
+ 		first_elem = g_list_next (first_elem); 
+	    }
+	}
+    }
+
+    if (!next_elem) {
+	if (priv->find_is_forward != forward) {
+	   g_list_reverse (next_list);
+	}
+
+	if (priv->find_str) {
+	    str = g_strdup (priv->find_str);
+
+	    if (!match_case) {
+		haystack = g_utf8_casefold (str, -1);
+	    } else {
+		haystack = g_strdup (str);
+	    }
+
+	    if (!strcmp (find_string, haystack)) {
+		box_node = priv->find_elem->data;
+
+		if (priv->find_is_forward) {
+		    html_find_selection_set (priv->view,
+					     box_node->find_box->dom_node,
+					     box_node->find_index + (priv->find_offset - 
+								     strlen (priv->find_str)),
+					     strlen (priv->find_str));
+		} else {
+		    html_find_selection_set (priv->view,
+					     box_node->find_box->dom_node,
+					     box_node->find_index + (box_node->find_len - 
+								     priv->find_offset),
+					     strlen (priv->find_str));		
+		}
+	    }
+	    
+	    g_free (str);
+	    g_free (haystack);
+	}
+    } else {
+	priv->find_list   = next_list;
+	priv->find_elem   = next_elem;	
+	priv->find_offset = next_offset;
+
+	g_free (priv->find_str);
+	priv->find_str = found_str;
+
+	priv->find_is_forward = forward;
+    }
+
+    return next_elem != NULL;
+}
+
+gboolean
 yelp_html_find (YelpHtml    *html,
 		const gchar *find_string,
 		gboolean     match_case,
@@ -405,158 +760,33 @@ yelp_html_find (YelpHtml    *html,
 		gboolean     forward)
 {
     YelpHtmlPriv    *priv;
-    DomNode         *root;
-    DomNode         *tmp_node;
     DomNode         *node;
-    DomNodeIterator *iter;
-    gchar           *str;
-    gchar           *hit;
-    gchar           *haystack;
 
-    HtmlBox         *box;
-    gchar           *box_text;
-    gint             len;
-
-    g_return_if_fail (YELP_IS_HTML (html));
-    g_return_if_fail (find_string != NULL);
+    g_return_val_if_fail (YELP_IS_HTML (html), FALSE);
+    g_return_val_if_fail (find_string != NULL, FALSE);
 
     priv = html->priv;
 
-    if (!priv->find_iter) {
-	iter = dom_DocumentTraversal_createNodeIterator
-	    (DOM_DOCUMENT_TRAVERSAL (priv->doc->dom_document),
-	     DOM_NODE (priv->doc->dom_document),
-	     DOM_SHOW_ALL,
-	     NULL,
-	     FALSE,
-	     NULL);
+    if (!priv->find_iter) { 
+	node = html_get_dom_node (priv->doc, "body");
 
-	root = NULL;
-	while ((node = dom_NodeIterator_nextNode (iter, NULL))) {
-	    if (!g_ascii_strcasecmp (dom_Node__get_nodeName (node), "body")) {
-		root = node;
-		break;
-	    }
-	}
-
-	g_object_unref (iter);
-	iter = NULL;
-
-	if (!root) {
-	    g_warning ("html_find(): Couldn't find html body.");
-	    return;
+	if (!node) {
+	    g_warning ("yelp_html_find(): Couldn't find html body.");  
+	    return FALSE;
 	}
 
 	priv->find_iter = dom_DocumentTraversal_createNodeIterator
 	    (DOM_DOCUMENT_TRAVERSAL (priv->doc->dom_document),
-	     root,
+	     node,
 	     DOM_SHOW_TEXT,
 	     NULL,
 	     FALSE,
 	     NULL);
 
-	if (forward) {
-	    priv->find_node = dom_NodeIterator_nextNode (priv->find_iter, NULL);
-	    priv->find_offset = 0;
-	} else {
-	    do {
-		tmp_node = priv->find_node;
-		priv->find_node = dom_NodeIterator_nextNode (priv->find_iter, NULL);
-	    } while (priv->find_node);
+	html_get_box_list (html);
 
-	    priv->find_node = tmp_node;
-
-	    if (priv->find_node) {
-		box = html_view_find_layout_box (priv->view, priv->find_node, FALSE);
-		box_text = html_box_text_get_text (HTML_BOX_TEXT (box), &len);
-		priv->find_offset = len;
-	    }
-	}
+	html_find_reset (html);
     }
 
-    while (priv->find_node) {
-	box = html_view_find_layout_box (priv->view, priv->find_node, FALSE);
-
-	/* We get the text from the layout box instead from the DOM node
-	 * directly, since the text in the box is canonicalized
-	 * (whitespace stripped in places etc).
-	 */
-	box_text = html_box_text_get_text (HTML_BOX_TEXT (box), &len);
-	if (len) {
-	    str = g_new (gchar, len + 1);
-	    memcpy (str, box_text, sizeof (gchar) * len);
-	    str[len] = 0;
-	} else {
-	    str = g_strdup ("");
-	}
-
-	if (!match_case) {
-	    if (forward)
-		haystack = g_utf8_casefold (str + priv->find_offset,
-					    len - priv->find_offset);
-	    else
-		haystack = g_utf8_casefold (str,
-					    priv->find_offset);
-	} else {
-	    if (forward)
-		haystack = g_strdup (str + priv->find_offset);
-	    else
-		haystack = g_strndup (str, priv->find_offset);
-	}
-
-	if (forward) {
-	    hit = strstr (haystack, find_string);
-	} else {
-	    hit = g_strrstr_len (haystack,
-				 priv->find_offset,
-				 find_string);
-	}
-
-	if (hit) {
-	    if (forward) {
-		html_selection_set (priv->view,
-				    priv->find_node,
-				    hit - haystack + priv->find_offset,
-				    strlen (find_string));
-		priv->find_offset += hit - haystack + strlen (find_string);
-	    } else {
-		html_selection_set (priv->view,
-				    priv->find_node,
-				    hit - haystack,
-				    strlen (find_string));
-		priv->find_offset = hit - haystack;
-	    }
-
-	    html_view_scroll_to_node (priv->view,
-				      priv->find_node,
-				      HTML_VIEW_SCROLL_TO_TOP);
-	}
-
-	g_free (str);
-	g_free (haystack);
-
-	if (hit) {
-	    break;
-	}
-
-	if (forward) {
-	    priv->find_node = dom_NodeIterator_nextNode (priv->find_iter, NULL);
-	    priv->find_offset = 0;
-	} else {
-	    priv->find_node = dom_NodeIterator_previousNode (priv->find_iter, NULL);
-	    if (priv->find_node) {
-		box = html_view_find_layout_box (priv->view, priv->find_node, FALSE);
-		box_text = html_box_text_get_text (HTML_BOX_TEXT (box), &len);
-		priv->find_offset = len;
-	    }
-	}			
-    }
-
-    if (!priv->find_node) {
-	g_object_unref (priv->find_iter);
-	priv->find_iter = NULL;
-	priv->find_offset = 0;
-	html_selection_clear (priv->view);
-    }
+    return html_find_text (html, find_string, match_case, wrap, forward);
 }
-
