@@ -1,5 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 4 -*- */
 /*
+ * Copyright (C) 2001-2002 Mikael Hallendal <micke@imendio.com>
  * Copyright (C) 2003 Shaun McCance <shaunm@gnome.org>
  *
  * This program is free software; you can redistribute it and/or
@@ -17,8 +18,8 @@
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  * Boston, MA 02111-1307, USA.
  *
+ * Author: Mikael Hallendal <micke@imendio.com>
  * Author: Shaun McCance <shaunm@gnome.org>
- *   Based on implementation by Mikael Hallendal <micke@imendio.com>
  */
 
 #ifdef HAVE_CONFIG_H
@@ -43,23 +44,29 @@
 #include "yelp-history.h"
 #include "yelp-html.h"
 #include "yelp-pager.h"
-#include "yelp-section.h"
 #include "yelp-toc-pager.h"
-#include "yelp-util.h"
+#include "yelp-theme.h"
 #include "yelp-window.h"
 
 #define d(x)
-#define RESPONSE_PREV 1
-#define RESPONSE_NEXT 2
-
 #define YELP_CONFIG_WIDTH  "/yelp/Geometry/width"
 #define YELP_CONFIG_HEIGHT "/yelp/Geometry/height"
 #define YELP_CONFIG_WIDTH_DEFAULT  "600"
 #define YELP_CONFIG_HEIGHT_DEFAULT "420"
 
 typedef enum {
-	YELP_WINDOW_ACTION_BACK = 1,
-	YELP_WINDOW_ACTION_FORWARD
+    YELP_WINDOW_FIND_PREV = 1,
+    YELP_WINDOW_FIND_NEXT
+} YelpFindAction;
+
+typedef enum {
+    YELP_WINDOW_GO_BACK = 1,
+    YELP_WINDOW_GO_FORWARD,
+    YELP_WINDOW_GO_PREVIOUS,
+    YELP_WINDOW_GO_NEXT,
+    YELP_WINDOW_GO_TOC,
+
+    YELP_WINDOW_GO_HOME
 } YelpHistoryAction;
 
 
@@ -73,16 +80,19 @@ static GtkWidget * window_create_toolbar          (YelpWindow        *window);
 static GdkPixbuf * window_load_icon               (void);
 static void        window_set_sections            (YelpWindow        *window,
 						   GtkTreeModel      *sections);
-static gboolean    window_handle_uri              (YelpWindow        *window,
+static void        window_handle_uri              (YelpWindow        *window,
 						   YelpURI           *uri);
 static gboolean    window_handle_pager_uri        (YelpWindow        *window,
 						   YelpURI           *uri);
 static gboolean    window_handle_html_uri         (YelpWindow        *window,
 						   YelpURI           *uri);
+static void        window_handle_page             (YelpWindow        *window,
+						   YelpPage          *page,
+						   YelpURI           *uri);
 static void        window_disconnect              (YelpWindow        *window);
 
-static gboolean    window_loading                 (YelpWindow        *window);
-
+static void        pager_contents_cb              (YelpPager         *pager,
+						   gpointer           user_data);
 static void        pager_page_cb                  (YelpPager         *pager,
 						   gchar             *page_id,
 						   gpointer           user_data);
@@ -128,44 +138,36 @@ static void        window_toggle_history_forward  (YelpHistory       *history,
 static void        window_history_action          (YelpWindow        *window,
 						   YelpHistoryAction  action);
 
-static void        window_history_go_cb           (gpointer           data,
-						   guint              section,
+static void        window_go_cb                   (gpointer           data,
+						   guint              action,
 						   GtkWidget         *widget);
 
-static void        window_go_home_cb              (gpointer           data,
-						   guint              section,
+static void        window_find_cb                 (gpointer           data,
+						   guint              action,
 						   GtkWidget         *widget);
-
-static void        window_find_cb                 (gpointer data,     guint section,
-						   GtkWidget         *widget);
-static void        window_find_again_cb           (gpointer data,     guint section,
+static void        window_find_again_cb           (gpointer           data,
+						   guint              action,
 						   GtkWidget         *widget);
+static gboolean    window_find_action             (YelpWindow        *window, 
+						   YelpFindAction     action);
 static gboolean    window_find_delete_event_cb    (GtkWidget         *widget,
 						   GdkEvent          *event,
  			                           gpointer           data);
+static void        window_find_entry_changed_cb   (GtkEditable       *editable,
+						   gpointer           data);
+static void        window_find_case_toggled_cb    (GtkWidget         *widget,
+						   gpointer           data);
+static void        window_find_wrap_toggled_cb    (GtkWidget         *widget,
+						   gpointer           data);
+static void        window_find_save_settings      (YelpWindow        *window);
+static void        window_find_buttons_set_sensitive (YelpWindow      *window,
+						      gboolean        sensitive);
 static void        window_find_response_cb        (GtkWidget         *dialog ,
 						   gint               response,
 						   YelpWindow        *window);
 
-/*
-static void        window_index_button_clicked    (GtkWidget         *button,
-						   YelpWindow        *window);
-static void        window_history_go_cb           (gpointer           data,
-						   guint              section,
-						   GtkWidget         *widget);
-static void        window_go_index_cb             (gpointer           data,
-						   guint              section,
-						   GtkWidget         *widget);
-static YelpView *  window_get_active_view         (YelpWindow        *window);
-*/
-
-/*
-enum {
-    PAGE_TOC_VIEW,
-    PAGE_CONTENT_VIEW,
-    PAGE_INDEX_VIEW
-};
-*/
+static gboolean    tree_model_iter_following      (GtkTreeModel      *model,
+						   GtkTreeIter       *iter);
 
 enum {
     NEW_WINDOW_REQUESTED,
@@ -190,6 +192,10 @@ struct _YelpWindowPriv {
     GtkWidget      *find_entry;
     GtkWidget      *case_checkbutton;
     GtkWidget      *wrap_checkbutton;
+    GtkWidget      *find_prev_button;
+    GtkWidget      *find_next_button;
+    GtkWidget      *find_close_button;
+
     gchar          *find_string;
     gboolean        match_case;
     gboolean        wrap;
@@ -197,6 +203,7 @@ struct _YelpWindowPriv {
     YelpHistory    *history;
 
     YelpPager      *pager;
+    gulong          contents_handler;
     gulong          page_handler;
     gulong          error_handler;
     gulong          finish_handler;
@@ -206,7 +213,10 @@ struct _YelpWindowPriv {
     GtkWidget      *forward_button;
     GtkWidget      *back_button;
 
-    gint            loading;
+    /* Don't free these */
+    gchar          *prev;
+    gchar          *next;
+    gchar          *toc;
 };
 
 static GtkItemFactoryEntry menu_items[] = {
@@ -219,28 +229,36 @@ static GtkItemFactoryEntry menu_items[] = {
      "<StockItem>",                  GTK_STOCK_CLOSE            },
 
     {N_("/_Edit"),                   NULL, 0, 0, "<Branch>"     },
-    {N_("/Edit/_Find in page..."),   NULL,
+    {N_("/Edit/_Find..."),           "<Control>f",
      window_find_cb,                 0,
      "<StockItem>",                  GTK_STOCK_FIND             },
-    {N_("/Edit/_Find again"),        "<Control>g",
-     window_find_again_cb,           0,
-     "<StockItem>",                  GTK_STOCK_FIND             },
+    {N_("/Edit/Find Ne_xt"),        "<Control>g",
+     window_find_again_cb,           YELP_WINDOW_FIND_NEXT,
+     "<Item>",                       NULL                       },
+    {N_("/Edit/Find Pre_vious"),    "<Shift><Control>g",
+     window_find_again_cb,           YELP_WINDOW_FIND_PREV, 
+     "<Item>",                       NULL                       },
 
     {N_("/_Go"),                     NULL, 0, 0, "<Branch>"     },
-    {N_("/Go/_Back"),                NULL,
-     window_history_go_cb,           YELP_WINDOW_ACTION_BACK,
+    {N_("/Go/_Back"),                "<Alt>Left",
+     window_go_cb,                   YELP_WINDOW_GO_BACK,
      "<StockItem>",                  GTK_STOCK_GO_BACK          },
-    {N_("/Go/_Forward"),             NULL,
-     window_history_go_cb,           YELP_WINDOW_ACTION_FORWARD,
+    {N_("/Go/_Forward"),             "<Alt>Right",
+     window_go_cb,                   YELP_WINDOW_GO_FORWARD,
      "<StockItem>",                  GTK_STOCK_GO_FORWARD       },
     {N_("/Go/_Home"),                NULL,
-     window_go_home_cb,              0,
+     window_go_cb,                   YELP_WINDOW_GO_HOME,
      "<StockItem>",                  GTK_STOCK_HOME             },
-    /*
-    {N_("/Go/_Index"),               NULL,
-     window_go_index_cb,             0,
-     "<StockItem>",                  GTK_STOCK_INDEX            },
-    */
+    {N_("/Go/"),                     NULL, 0, 0, "<Separator>"  },
+    {N_("/Go/_Previous"),            "<Alt>Up",
+     window_go_cb,                   YELP_WINDOW_GO_PREVIOUS,
+     0,                              NULL                       },
+    {N_("/Go/_Next"),                "<Alt>Down",
+     window_go_cb,                   YELP_WINDOW_GO_NEXT,
+     0,                              NULL                       },
+    {N_("/Go/_Contents"),            NULL,
+     window_go_cb,                   YELP_WINDOW_GO_TOC,
+     0,                              NULL                       },
 
     {N_("/_Help"),                   NULL, 0, 0, "<Branch>"     },
     {N_("/Help/_About"),             NULL,
@@ -348,13 +366,88 @@ void
 yelp_window_open_uri (YelpWindow  *window,
 		      YelpURI     *uri)
 {
+    YelpWindowPriv *priv;
+    GError         *error = NULL;
+    YelpURI        *cur_uri;
+
     g_return_if_fail (YELP_IS_WINDOW (window));
+    g_return_if_fail (YELP_IS_URI (uri));
+
+    priv = window->priv;
+
+    cur_uri = yelp_history_get_current (window->priv->history);
+    if (cur_uri && yelp_uri_equal (cur_uri, uri))
+ 	return;
 
     yelp_history_goto (window->priv->history, uri);
 
+    if (!yelp_uri_exists (uri)) {
+	gchar *str_uri = yelp_uri_to_string (uri);
+		
+	g_set_error (&error,
+		     YELP_ERROR,
+		     YELP_ERROR_URI_NOT_EXIST,
+		     _("The document '%s' does not exist"), str_uri);
+	window_error (window, error);
+
+	g_free (str_uri);
+	g_error_free (error);
+
+	return;
+    }
+
     window_handle_uri (window, uri);
 }
 
+void
+window_handle_uri (YelpWindow  *window,
+		   YelpURI     *uri)
+{
+    YelpWindowPriv *priv;
+    GError         *error = NULL;
+    gboolean        handled = FALSE;
+
+    g_return_if_fail (YELP_IS_WINDOW (window));
+    g_return_if_fail (YELP_IS_URI (uri));
+
+    priv = window->priv;
+
+    switch (yelp_uri_get_resource_type (uri)) {
+    case YELP_URI_TYPE_DOCBOOK_XML:
+    case YELP_URI_TYPE_MAN:
+    case YELP_URI_TYPE_INFO:
+    case YELP_URI_TYPE_TOC:
+	handled = window_handle_pager_uri (window, uri);
+	break;
+    case YELP_URI_TYPE_DOCBOOK_SGML:
+	g_set_error (&error,
+		     YELP_ERROR,
+		     YELP_ERROR_NO_SGML,
+		     _("DocBook SGML documents are no longer supported."));
+
+	window_error (window, error);
+	g_error_free (error);
+	return;
+    case YELP_URI_TYPE_HTML:
+	handled = window_handle_html_uri (window, uri);
+	break;
+    case YELP_URI_TYPE_GHELP:
+    case YELP_URI_TYPE_GHELP_OTHER:
+    case YELP_URI_TYPE_INDEX:
+    case YELP_URI_TYPE_PATH:
+    case YELP_URI_TYPE_FILE:
+    case YELP_URI_TYPE_UNKNOWN:
+    case YELP_URI_TYPE_RELATIVE:
+    default:
+	break;
+    }
+ 
+    if (error)
+	window_error (window, error);
+
+    window_find_buttons_set_sensitive (window, TRUE);    
+}
+
 YelpURI *
 yelp_window_get_current_uri (YelpWindow *window)
 {
@@ -414,13 +507,13 @@ window_populate (YelpWindow *window)
 
     menu_item =
 	gtk_item_factory_get_item_by_action (priv->item_factory,
-					     YELP_WINDOW_ACTION_BACK);
+					     YELP_WINDOW_GO_BACK);
     if (menu_item)
 	gtk_widget_set_sensitive (menu_item, FALSE);
 
     menu_item =
 	gtk_item_factory_get_item_by_action (priv->item_factory,
-					     YELP_WINDOW_ACTION_FORWARD);
+					     YELP_WINDOW_GO_FORWARD);
     if (menu_item)
 	gtk_widget_set_sensitive (menu_item, FALSE);
 
@@ -541,18 +634,16 @@ window_load_icon (void)
     static GdkPixbuf *pixbuf = NULL;
 
     if (!pixbuf) {
-	gchar *file;
+	GError *error = NULL;
+	GtkIconTheme *icon_theme = (GtkIconTheme *) yelp_theme_get_icon_theme ();
 
-	file = gnome_program_locate_file (NULL,
-					  GNOME_FILE_DOMAIN_PIXMAP,
-					  "gnome-help.png",
-					  TRUE,
-					  NULL);
-
-	if (file) {
-	    pixbuf = gdk_pixbuf_new_from_file (file,
-					       NULL);
-	    g_free (file);
+	pixbuf = gtk_icon_theme_load_icon (icon_theme,
+					   "gnome-help",
+					   36, 0,
+					   &error);
+	if (!pixbuf) {
+	    g_warning ("Couldn't load icon: %s", error->message);
+	    g_error_free (error);
 	}
     }
 
@@ -613,70 +704,6 @@ window_set_sections (YelpWindow   *window,
     }
 }
 
-static gboolean
-window_handle_uri (YelpWindow *window,
-		   YelpURI    *uri)
-{
-    YelpWindowPriv *priv;
-    GError         *error = NULL;
-    gboolean        handled = FALSE;
-
-    priv = window->priv;
-
-    g_return_val_if_fail (YELP_IS_WINDOW (window), FALSE);
-    g_return_val_if_fail (YELP_IS_URI (uri), FALSE);
-
-    if (!yelp_uri_exists (uri)) {
-	gchar *str_uri = yelp_uri_to_string (uri);
-		
-	g_set_error (&error,
-		     YELP_ERROR,
-		     YELP_ERROR_URI_NOT_EXIST,
-		     _("The document '%s' does not exist"), str_uri);
-	window_error (window, error);
-
-	g_free (str_uri);
-	g_error_free (error);
-
-	return FALSE;
-    }
-
-    switch (yelp_uri_get_resource_type (uri)) {
-    case YELP_URI_TYPE_DOCBOOK_XML:
-    case YELP_URI_TYPE_MAN:
-    case YELP_URI_TYPE_INFO:
-    case YELP_URI_TYPE_TOC:
-	handled = window_handle_pager_uri (window, uri);
-	break;
-    case YELP_URI_TYPE_DOCBOOK_SGML:
-	g_set_error (&error,
-		     YELP_ERROR,
-		     YELP_ERROR_NO_SGML,
-		     _("DocBook SGML documents are no longer supported."));
-
-	window_error (window, error);
-	g_error_free (error);
-	return FALSE;
-    case YELP_URI_TYPE_HTML:
-	handled = window_handle_html_uri (window, uri);
-	break;
-    case YELP_URI_TYPE_GHELP:
-    case YELP_URI_TYPE_GHELP_OTHER:
-    case YELP_URI_TYPE_INDEX:
-    case YELP_URI_TYPE_PATH:
-    case YELP_URI_TYPE_FILE:
-    case YELP_URI_TYPE_UNKNOWN:
-    case YELP_URI_TYPE_RELATIVE:
-    default:
-	break;
-    }
-
-    if (error)
-	window_error (window, error);
-
-    return handled;
-}
-
 static gboolean
 window_handle_pager_uri (YelpWindow *window,
 			 YelpURI    *uri)
@@ -689,6 +716,7 @@ window_handle_pager_uri (YelpWindow *window,
     gchar      *path;
     YelpPage   *page = NULL;
     YelpPager  *pager;
+    YelpPagerState state;
 
     priv = window->priv;
 
@@ -738,40 +766,111 @@ window_handle_pager_uri (YelpWindow *window,
 	g_object_unref (priv->pager);
     priv->pager = pager;
 
-    switch (yelp_pager_get_state (pager)) {
-    case YELP_PAGER_STATE_START:
+    state = yelp_pager_get_state (pager);
+
+    loadnow  = FALSE;
+    startnow = TRUE;
+
+    if (state & YELP_PAGER_STATE_STOPPED) {
+	error = yelp_pager_get_error (pager);
+	if (error) {
+	    window_error (window, error);
+	    return FALSE;
+	}
+    }
+
+    if (state & YELP_PAGER_STATE_STARTED) {
+	if (state & YELP_PAGER_STATE_CONTENTS) {
+	    gchar *frag_id = yelp_uri_get_fragment (uri);
+	    gchar *page_id = yelp_pager_resolve_uri (pager, uri);
+
+	    if (!page_id && (frag_id && strcmp (frag_id, ""))) {
+		g_set_error (&error,
+			     YELP_ERROR,
+			     YELP_ERROR_FAILED_OPEN,
+			     _("The page '%s' could not be found in this document."),
+			     frag_id);
+		window_error (window, error);
+
+		g_free (frag_id);
+		g_free (page_id);
+		g_error_free (error);
+
+		return FALSE;
+	    }
+
+	    g_free (frag_id);
+	    g_free (page_id);
+	} else {
+	    priv->contents_handler =
+		g_signal_connect (pager,
+				  "contents",
+				  G_CALLBACK (pager_contents_cb),
+				  window);
+	}
+
 	page = (YelpPage *) yelp_pager_lookup_page (pager, uri);
 	loadnow  = (page ? TRUE : FALSE);
 	startnow = FALSE;
-	break;
-    case YELP_PAGER_STATE_NEW:
-    case YELP_PAGER_STATE_CANCEL:
-	loadnow  = FALSE;
-	startnow = TRUE;
-	break;
-    case YELP_PAGER_STATE_FINISH:
-	page = (YelpPage *) yelp_pager_lookup_page (pager, uri);
-	loadnow = TRUE;
-	break;
-    case YELP_PAGER_STATE_ERROR:
-	error = yelp_pager_get_error (pager);
-	window_error (window, error);
-	return FALSE;
-    default:
-    	g_assert_not_reached ();
-	break;
+
+	if (state & YELP_PAGER_STATE_FINISHED) {
+	    if (!page) {
+		gchar *frag_id = yelp_uri_get_fragment (uri);
+		g_set_error (&error,
+			     YELP_ERROR,
+			     YELP_ERROR_FAILED_OPEN,
+			     _("The page '%s' could not be found in this document."),
+			     frag_id);
+		window_error (window, error);
+
+		g_free (frag_id);
+		g_error_free (error);
+		return FALSE;
+	    }
+	}
+    } else {
+	priv->contents_handler =
+	    g_signal_connect (pager,
+			      "contents",
+			      G_CALLBACK (pager_contents_cb),
+			      window);
     }
 
-    if (!loadnow) {
+    window_set_sections (window,
+			 GTK_TREE_MODEL (yelp_pager_get_sections (pager)));
+
+    if (loadnow)
+	window_handle_page (window, page, uri);
+
+    else {
+	GtkWidget *menu_item;
 	gchar *loading = _("Loading...");
+	GdkCursor *cursor = gdk_cursor_new (GDK_WATCH);
+
 	yelp_html_clear (priv->html_view);
 
+	gdk_window_set_cursor (GTK_WIDGET (window)->window, cursor);
+	gdk_cursor_unref (cursor);
+
+	menu_item =
+	    gtk_item_factory_get_item_by_action (priv->item_factory,
+						 YELP_WINDOW_GO_PREVIOUS);
+	if (menu_item)
+	    gtk_widget_set_sensitive (menu_item, FALSE);
+	menu_item =
+	    gtk_item_factory_get_item_by_action (priv->item_factory,
+						 YELP_WINDOW_GO_NEXT);
+	if (menu_item)
+	    gtk_widget_set_sensitive (menu_item, FALSE);
+	menu_item =
+	    gtk_item_factory_get_item_by_action (priv->item_factory,
+						 YELP_WINDOW_GO_TOC);
+	if (menu_item)
+	    gtk_widget_set_sensitive (menu_item, FALSE);
+
 	gtk_window_set_title (GTK_WINDOW (window),
 			      (const gchar *) loading);
 
-	window_set_sections (window,
-			     GTK_TREE_MODEL (yelp_pager_get_sections (pager)));
-
 	yelp_html_printf
 	    (priv->html_view,
 	     "<html><head><meta http-equiv='Content-Type'"
@@ -782,9 +881,6 @@ window_handle_pager_uri (YelpWindow *window,
 	     loading, loading);
 	yelp_html_close (priv->html_view);
 
-	priv->loading = 1;
-	gtk_timeout_add (200, (GtkFunction) window_loading, window);
-
 	yelp_toc_pager_pause (yelp_toc_pager_get ());
 	while (gtk_events_pending ())
 	    gtk_main_iteration ();
@@ -808,35 +904,6 @@ window_handle_pager_uri (YelpWindow *window,
 
 	if (startnow)
 	    yelp_pager_start (pager);
-    } else {
-	if (!page) {
-	    gchar *str_uri = yelp_uri_to_string (uri);
-	    g_set_error (&error,
-			 YELP_ERROR,
-			 YELP_ERROR_FAILED_OPEN,
-			 _("The document '%s' could not be opened"), str_uri);
-
-	    window_error (window, error);
-
-	    g_free (str_uri);
-	    g_error_free (error);
-
-	    return FALSE;
-	}
-
-	window_set_sections (window,
-			     GTK_TREE_MODEL (yelp_pager_get_sections (pager)));
-
-	yelp_html_clear (priv->html_view);
-	yelp_html_set_base_uri (priv->html_view, uri);
-
-	gtk_window_set_title (GTK_WINDOW (window),
-			      (const gchar *) page->title);
-
-	priv->loading = 0;
-	yelp_html_write (priv->html_view,
-			 page->chunk,
-			 strlen (page->chunk));
     }
 
     return TRUE;
@@ -849,13 +916,95 @@ window_handle_html_uri (YelpWindow    *window,
     return FALSE;
 }
 
+static void
+window_handle_page (YelpWindow   *window,
+		    YelpPage     *page,
+		    YelpURI      *uri)
+{
+    GtkWidget      *menu_item;
+    GtkTreeModel   *model;
+    GtkTreeIter     iter;
+    YelpWindowPriv *priv;
+    gchar          *id;
+    gboolean        valid;
+
+    g_return_if_fail (YELP_IS_WINDOW (window));
+    g_return_if_fail (YELP_IS_URI (uri));
+
+    priv = window->priv;
+
+    model = gtk_tree_view_get_model (GTK_TREE_VIEW (priv->side_sects));
+
+    if (model) {
+	valid = gtk_tree_model_get_iter_first (model, &iter);
+	while (valid) {
+	    gtk_tree_model_get (model, &iter,
+				0, &id,
+				-1);
+	    if (yelp_pager_uri_is_page (priv->pager, id, uri)) {
+		GtkTreePath *path = gtk_tree_model_get_path (model, &iter);
+
+		gtk_tree_view_expand_to_path (GTK_TREE_VIEW (priv->side_sects),
+					      path);
+		gtk_tree_view_set_cursor (GTK_TREE_VIEW (priv->side_sects),
+					  path, NULL, FALSE);
+
+		gtk_tree_path_free (path);
+		break;
+	    }
+
+	    valid = tree_model_iter_following (model, &iter);
+	}
+    }
+
+    if (page->prev) {
+	priv->prev = page->prev;
+	menu_item =
+	    gtk_item_factory_get_item_by_action (priv->item_factory,
+						 YELP_WINDOW_GO_PREVIOUS);
+	if (menu_item)
+	    gtk_widget_set_sensitive (menu_item, TRUE);
+    }
+    if (page->next) {
+	priv->next = page->next;
+	menu_item =
+	    gtk_item_factory_get_item_by_action (priv->item_factory,
+						 YELP_WINDOW_GO_NEXT);
+	if (menu_item)
+	    gtk_widget_set_sensitive (menu_item, TRUE);
+    }
+    if (page->toc) {
+	priv->toc = page->toc;
+	menu_item =
+	    gtk_item_factory_get_item_by_action (priv->item_factory,
+						 YELP_WINDOW_GO_TOC);
+	if (menu_item)
+	    gtk_widget_set_sensitive (menu_item, TRUE);
+    }
+
+    gtk_window_set_title (GTK_WINDOW (window),
+			  (const gchar *) page->title);
+
+    yelp_html_clear (priv->html_view);
+    yelp_html_set_base_uri (priv->html_view, uri);
+    yelp_html_write (priv->html_view,
+		     page->chunk,
+		     strlen (page->chunk));
+}
+
 static void
 window_disconnect (YelpWindow *window)
 {
     YelpWindowPriv *priv = window->priv;
     g_return_if_fail (YELP_IS_WINDOW (window));
 
-    priv->loading = 0;
+    gdk_window_set_cursor (GTK_WIDGET (window)->window, NULL);
+
+    if (priv->contents_handler) {
+	g_signal_handler_disconnect (priv->pager,
+				     priv->contents_handler);
+	priv->contents_handler = 0;
+    }
     if (priv->page_handler) {
 	g_signal_handler_disconnect (priv->pager,
 				     priv->page_handler);
@@ -873,45 +1022,33 @@ window_disconnect (YelpWindow *window)
     }
 }
 
-static gboolean
-window_loading (YelpWindow *window)
+static void
+pager_contents_cb (YelpPager   *pager,
+		   gpointer     user_data)
 {
-    gchar *loading;
-    YelpWindowPriv *priv = window->priv;
+    YelpWindow *window = YELP_WINDOW (user_data);
+    YelpURI    *uri;
+    GError     *error = NULL;
 
-    switch (priv->loading) {
-    case 0:
-	return FALSE;
-    case 1:
-	priv->loading = 2;
-	loading = _("Loading&nbsp;&nbsp;&nbsp;");
-	break;
-    case 2:
-	priv->loading = 3;
-	loading = _("Loading.&nbsp;&nbsp;");
-	break;
-    case 3:
-	priv->loading = 4;
-	loading = _("Loading..&nbsp;");
-	break;
-    default:
-	priv->loading = 1;
-	loading = _("Loading...");
-	break;
+    uri  = yelp_window_get_current_uri (window);
+
+    gchar *frag = yelp_uri_get_fragment (uri);
+    gchar *page = yelp_pager_resolve_uri (pager, uri);
+
+    if (!page && (frag && strcmp (frag, ""))) {
+	gchar *frag_id = yelp_uri_get_fragment (uri);
+	g_set_error (&error,
+		     YELP_ERROR,
+		     YELP_ERROR_FAILED_OPEN,
+		     _("The page '%s' could not be found in this document."),
+		     frag_id);
+	window_error (window, error);
+
+	g_free (frag_id);
+	g_error_free (error);
+
+	window_disconnect (window);
     }
-
-    yelp_html_clear (priv->html_view);
-    yelp_html_printf
-	(priv->html_view,
-	 "<html><head><meta http-equiv='Content-Type'"
-	 " content='text/html=; charset=utf-8'>"
-	 "<title>%s</title></head>"
-	 "<body><center>%s</center></body>"
-	 "</html>",
-	 loading, loading);
-    yelp_html_close (priv->html_view);
-
-    return TRUE;
 }
 
 static void
@@ -927,17 +1064,9 @@ pager_page_cb (YelpPager *pager,
 
     if (yelp_pager_uri_is_page (pager, page_id, uri)) {
 	window_disconnect (window);
+
 	page = (YelpPage *) yelp_pager_get_page (pager, page_id);
-
-	yelp_html_clear (window->priv->html_view);
-	yelp_html_set_base_uri (window->priv->html_view, uri);
-
-	gtk_window_set_title (GTK_WINDOW (window),
-			      (const gchar *) page->title);
-
-	yelp_html_write (window->priv->html_view,
-			 page->chunk,
-			 strlen (page->chunk));
+	window_handle_page (window, page, uri);
     }
 }
 
@@ -963,21 +1092,21 @@ pager_finish_cb (YelpPager   *pager,
     GError *error = NULL;
     YelpWindow *window = YELP_WINDOW (user_data);
     YelpURI *uri;
-    gchar   *str_uri;
+    gchar   *frag_id;
 
     uri     = yelp_window_get_current_uri (window);
-    str_uri = yelp_uri_to_string (uri);
+    frag_id = yelp_uri_get_fragment (uri);
 
     window_disconnect (window);
 
     g_set_error (&error,
 		 YELP_ERROR,
 		 YELP_ERROR_FAILED_OPEN,
-		 _("The document '%s' could not be opened"), str_uri);
-
+		 _("The page '%s' could not be found in this document."),
+		 frag_id);
     window_error (window, error);
 
-    g_free (str_uri);
+    g_free (frag_id);
     g_error_free (error);
 
     // FIXME: Remove the URI from the history and go back
@@ -1095,7 +1224,7 @@ window_back_clicked (GtkWidget  *button,
 {
     g_return_if_fail (YELP_IS_WINDOW (window));
 
-    window_history_action (window, YELP_WINDOW_ACTION_BACK);
+    window_history_action (window, YELP_WINDOW_GO_BACK);
 }
 
 static void
@@ -1104,30 +1233,53 @@ window_forward_clicked (GtkWidget  *button,
 {
     g_return_if_fail (YELP_IS_WINDOW (window));
 
-    window_history_action (window, YELP_WINDOW_ACTION_FORWARD);
+    window_history_action (window, YELP_WINDOW_GO_FORWARD);
 }
 
 static void        
-window_history_go_cb (gpointer   data,
-	  	      guint      section,
-	  	      GtkWidget *widget)
+window_go_cb (gpointer   data,
+	      guint      action,
+	      GtkWidget *widget)
 {
+    YelpWindow *window;
+    YelpURI *uri;
+
     g_return_if_fail (YELP_IS_WINDOW (YELP_WINDOW (data)));
 
-    window_history_action (YELP_WINDOW (data), section);    
+    window = YELP_WINDOW (data);
+
+    switch (action) {
+    case YELP_WINDOW_GO_BACK:
+    case YELP_WINDOW_GO_FORWARD:
+	window_history_action (YELP_WINDOW (data), action);
+	break;
+    case YELP_WINDOW_GO_HOME:
+	window_home_button_clicked (NULL, YELP_WINDOW (data));
+	break;
+
+    case YELP_WINDOW_GO_PREVIOUS:
+	uri = yelp_uri_new_relative (yelp_window_get_current_uri (window),
+				     window->priv->prev);
+	yelp_window_open_uri (window, uri);
+	g_object_unref (uri);
+	break;
+    case YELP_WINDOW_GO_NEXT:
+	uri = yelp_uri_new_relative (yelp_window_get_current_uri (window),
+				     window->priv->next);
+	yelp_window_open_uri (window, uri);
+	g_object_unref (uri);
+	break;
+    case YELP_WINDOW_GO_TOC:
+	uri = yelp_uri_new_relative (yelp_window_get_current_uri (window),
+				     window->priv->toc);
+	yelp_window_open_uri (window, uri);
+	g_object_unref (uri);
+	break;
+    default:
+	break;
+    }
 }
 
-static void
-window_go_home_cb (gpointer   data,
-		   guint      section,
-		   GtkWidget *widget)
-{
-    g_return_if_fail (YELP_IS_WINDOW (YELP_WINDOW (data)));
-
-    window_home_button_clicked (NULL, YELP_WINDOW (data));
-}
-
-
 static void
 window_home_button_clicked (GtkWidget  *button,
 			    YelpWindow *window)
@@ -1159,7 +1311,7 @@ window_toggle_history_back (YelpHistory *history,
     gtk_widget_set_sensitive (priv->back_button, sensitive);
 
     menu_item = gtk_item_factory_get_item_by_action (priv->item_factory,
-						     YELP_WINDOW_ACTION_BACK);
+						     YELP_WINDOW_GO_BACK);
     if (menu_item)
 	gtk_widget_set_sensitive (menu_item, sensitive);
 }
@@ -1180,7 +1332,7 @@ window_toggle_history_forward (YelpHistory *history,
     gtk_widget_set_sensitive (priv->forward_button, sensitive);
 
     menu_item = gtk_item_factory_get_item_by_action (priv->item_factory,
-						     YELP_WINDOW_ACTION_FORWARD);
+						     YELP_WINDOW_GO_FORWARD);
     if (menu_item)
 	gtk_widget_set_sensitive (menu_item, sensitive);
 }
@@ -1197,10 +1349,10 @@ window_history_action (YelpWindow        *window,
     priv = window->priv;
 
     switch (action) {
-    case YELP_WINDOW_ACTION_BACK:
+    case YELP_WINDOW_GO_BACK:
 	uri = yelp_history_go_back (priv->history);
 	break;
-    case YELP_WINDOW_ACTION_FORWARD:
+    case YELP_WINDOW_GO_FORWARD:
 	uri = yelp_history_go_forward (priv->history);
 	break;
     default:
@@ -1222,25 +1374,58 @@ window_find_delete_event_cb (GtkWidget *widget,
     return TRUE;
 }
 
-static void 
-window_find_again_cb (gpointer   data,
-                      guint      section,
-		      GtkWidget *widget)
+static void
+window_find_save_settings (YelpWindow * window)
+{
+    YelpWindowPriv *priv;
+    const gchar    *tmp;
+
+    g_return_if_fail (YELP_IS_WINDOW (window));
+
+    priv = window->priv;
+
+    tmp = gtk_entry_get_text (GTK_ENTRY (priv->find_entry));
+
+    priv->match_case = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (priv->case_checkbutton));
+    priv->wrap       = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (priv->wrap_checkbutton));
+
+    g_free (priv->find_string);
+		
+    if (!priv->match_case) {
+	priv->find_string = g_utf8_casefold (tmp, -1);
+    } else {
+	priv->find_string = g_strdup (tmp);
+    }				  
+}
+
+static gboolean
+window_find_action (YelpWindow *window, YelpFindAction action)
 {
     YelpWindowPriv *priv;
 
+    priv = window->priv;
+
+    window_find_save_settings (window);
+
+    return yelp_html_find (priv->html_view,
+			   priv->find_string,
+			   priv->match_case,
+			   priv->wrap,
+			   (action == YELP_WINDOW_FIND_NEXT) ? TRUE: FALSE);
+}
+
+static void 
+window_find_again_cb (gpointer   data,
+                      guint      action,
+		      GtkWidget *widget)
+{
+    YelpWindow *window;
 
     g_return_if_fail (YELP_IS_WINDOW (data));
-    
-    priv = YELP_WINDOW (data)->priv;
+ 
+    window = YELP_WINDOW (data);
 
-    if (priv->find_string) {
-	yelp_html_find (priv->html_view,
-			priv->find_string,
-			priv->match_case,
-			priv->wrap,
-			TRUE);
-    }
+    window_find_action (window, (YelpFindAction)action);
 }
 
 static void
@@ -1249,7 +1434,6 @@ window_find_response_cb (GtkWidget  *dialog,
 			 YelpWindow *window)
 {
     YelpWindowPriv * priv;
-    const gchar *tmp;
 
     priv = window->priv;
 
@@ -1258,26 +1442,20 @@ window_find_response_cb (GtkWidget  *dialog,
 	gtk_widget_hide (dialog);
 	break;
 	
-    case RESPONSE_PREV:
-    case RESPONSE_NEXT:
-	tmp = gtk_entry_get_text (GTK_ENTRY (priv->find_entry));
-	
-	priv->match_case = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (priv->case_checkbutton));
-	priv->wrap = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (priv->wrap_checkbutton));
-	
-	g_free (priv->find_string);
-
-	if (!priv->match_case) {
-	    priv->find_string = g_utf8_casefold (tmp, -1);
+    case YELP_WINDOW_FIND_NEXT:
+    case YELP_WINDOW_FIND_PREV:
+	if (!window_find_action (window, response)) {
+	    if (YELP_WINDOW_FIND_NEXT == response) {
+		gtk_widget_set_sensitive (priv->find_next_button, FALSE);
+		gtk_widget_set_sensitive (priv->find_prev_button, TRUE);
+	    } else {
+		gtk_widget_set_sensitive (priv->find_next_button, TRUE);
+		gtk_widget_set_sensitive (priv->find_prev_button, FALSE);
+	    }
 	} else {
-	    priv->find_string = g_strdup (tmp);
+	    window_find_buttons_set_sensitive (window, TRUE);
 	}
 
-	yelp_html_find (priv->html_view,
-			priv->find_string,
-			priv->match_case,
-			priv->wrap,
-			RESPONSE_NEXT == response);
 	break;
 
     default:
@@ -1285,14 +1463,83 @@ window_find_response_cb (GtkWidget  *dialog,
     }
 }
 
+static void
+window_find_entry_changed_cb (GtkEditable *editable,
+			      gpointer     data)
+{
+    YelpWindow     *window;
+    gchar          *text;
+
+    g_return_if_fail (YELP_IS_WINDOW(data));
+	
+    window = YELP_WINDOW (data);
+
+    text = gtk_editable_get_chars (editable, 0, -1);
+
+    if (!strlen (text)) {
+	window_find_buttons_set_sensitive (window, FALSE);
+    } else {
+	window_find_buttons_set_sensitive (window, TRUE);
+    }
+ 
+    g_free (text);
+}
+
+static void
+window_find_case_toggled_cb (GtkWidget *widget,
+			     gpointer   data)
+{
+    YelpWindow     *window;
+    YelpWindowPriv *priv;
+
+    g_return_if_fail (YELP_IS_WINDOW (data));
+
+    window = YELP_WINDOW (data);
+    priv   = window->priv;
+
+    window_find_entry_changed_cb (GTK_EDITABLE (priv->find_entry),
+				  window);
+}
+
+static void
+window_find_wrap_toggled_cb (GtkWidget *widget,
+			     gpointer   data)
+{
+    YelpWindow     *window;
+    YelpWindowPriv *priv;
+
+    g_return_if_fail (YELP_IS_WINDOW (data));
+
+    window = YELP_WINDOW (data);
+    priv   = window->priv;
+
+    window_find_entry_changed_cb (GTK_EDITABLE (priv->find_entry),
+				  window);
+}
+
+static void
+window_find_buttons_set_sensitive (YelpWindow *window,
+				  gboolean    sensitive)
+{
+    YelpWindowPriv *priv;
+
+    priv = window->priv;
+
+    if (priv->find_dialog) {
+	gtk_widget_set_sensitive (priv->find_next_button, sensitive);
+	gtk_widget_set_sensitive (priv->find_prev_button, sensitive);
+    }
+}
+
 static void
 window_find_cb (gpointer   data,     
-		guint      section,
+		guint      action,
 		GtkWidget *widget)
 {
     YelpWindow     *window;
     YelpWindowPriv *priv;
     GladeXML       *glade;
+    GtkWidget      *gnome_entry;
 
     g_return_if_fail (YELP_IS_WINDOW (data));
     
@@ -1308,14 +1555,37 @@ window_find_cb (gpointer   data,
 	    return;
 	}
 
-	priv->find_dialog = glade_xml_get_widget (glade, "find_dialog");	
-	priv->find_entry = glade_xml_get_widget (glade, "find_entry");
-	priv->case_checkbutton = glade_xml_get_widget (glade, "case_check");
-	priv->wrap_checkbutton = glade_xml_get_widget (glade, "wrap_check");
+	priv->find_dialog       = glade_xml_get_widget (glade, "find_dialog");	
+	priv->find_entry        = glade_xml_get_widget (glade, "find_entry");
+	priv->case_checkbutton  = glade_xml_get_widget (glade, "case_check");
+	priv->wrap_checkbutton  = glade_xml_get_widget (glade, "wrap_check");
+	priv->find_next_button  = glade_xml_get_widget (glade, "next_button");
+	priv->find_prev_button  = glade_xml_get_widget (glade, "prev_button");
+	priv->find_close_button = glade_xml_get_widget (glade, "close_button");	
+	gnome_entry             = glade_xml_get_widget (glade, "find_gnome_entry");
+
+	gtk_window_set_transient_for (GTK_WINDOW (priv->find_dialog), 
+				      GTK_WINDOW (window));
+	gtk_window_set_destroy_with_parent (GTK_WINDOW (priv->find_dialog),
+					    TRUE);
+
+	gtk_combo_set_use_arrows (GTK_COMBO (gnome_entry), FALSE);
 	
 	gtk_entry_set_activates_default (GTK_ENTRY (priv->find_entry), TRUE);
 	gtk_widget_grab_default (glade_xml_get_widget (glade, "next_button"));
 
+	gtk_widget_set_sensitive (priv->find_next_button, FALSE);
+	gtk_widget_set_sensitive (priv->find_prev_button, FALSE);
+
+	g_signal_connect (G_OBJECT (priv->find_entry), "changed",
+			  G_CALLBACK (window_find_entry_changed_cb), window);	
+
+	g_signal_connect (G_OBJECT (priv->wrap_checkbutton), "toggled",
+			  G_CALLBACK (window_find_wrap_toggled_cb), window);
+
+	g_signal_connect (G_OBJECT (priv->case_checkbutton), "toggled",
+			  G_CALLBACK (window_find_case_toggled_cb), window);
+
 	g_signal_connect (G_OBJECT (priv->find_dialog), "delete_event",
 			  G_CALLBACK (window_find_delete_event_cb), NULL);
 
@@ -1325,5 +1595,39 @@ window_find_cb (gpointer   data,
 	g_object_unref (glade);
     }
 
+    gtk_editable_select_region (GTK_EDITABLE (priv->find_entry), 0, -1);
+
     gtk_window_present (GTK_WINDOW (priv->find_dialog));
 }
+
+// This would be nice to have in GTK+
+static gboolean
+tree_model_iter_following (GtkTreeModel  *model,
+			   GtkTreeIter   *iter)
+{
+    gboolean     valid;
+    GtkTreeIter *old_iter = gtk_tree_iter_copy (iter);
+
+    if (gtk_tree_model_iter_has_child (model, iter))
+	return gtk_tree_model_iter_children (model, iter, old_iter);
+    else do {
+	valid = gtk_tree_model_iter_next (model, iter);
+
+	if (valid) {
+	    gtk_tree_iter_free (old_iter);
+	    return TRUE;
+	} else {
+	    *iter = *old_iter;
+
+	    valid = gtk_tree_model_iter_parent (model, iter, old_iter);
+
+	    if (!valid) {
+		gtk_tree_iter_free (old_iter);
+		return FALSE;
+	    }
+	}
+    } while (TRUE);
+
+    g_assert_not_reached ();
+    return FALSE;
+}
